#include <avformat.h>

#include "input.h"
#include "decode.h"
#include <float.h>

#define N 100000

static int fsizes[N];
static double ftimes[N];
static uint8_t ftypes[N];

void chunkise_ipb(int n)
{
  int i, chunk_id = 0;
  int p_chunk_dimension = 0, b_chunk_dimension = 0;
  int p_chunk_size = 0, b_chunk_size = 0;
  double p_min_time = 0, b_min_time = 0;
  double p_max_time = 0, b_max_time = 0;

  for (i = 0; i < n; i++) {
    int j;
    int b_chunk_frame[50], p_chunk_frame[50];

    switch (ftypes[i]) {
      case 1:
        /* I Type */
        if (p_chunk_dimension) {
          printf("%d\t%f\t%f\t%d\t1\t2 :\t", chunk_id++, p_min_time, p_max_time, p_chunk_size);
          for (j = 0; j < p_chunk_dimension; j++) {
            printf("%d ", p_chunk_frame[j]);
          }
          printf("\n");
        }
        if (b_chunk_dimension) {
          printf("%d\t%f\t%f\t%d\t1\t1 :\t", chunk_id++, b_min_time, b_max_time, b_chunk_size);
          for (j = 0; j < b_chunk_dimension; j++) {
            printf("%d ", b_chunk_frame[j]);
          }
          printf("\n");
        }
        p_chunk_size = b_chunk_size = 0;
        p_chunk_dimension = b_chunk_dimension = 0;
        p_min_time = b_min_time = DBL_MAX;
        p_max_time = b_max_time = 0;
        printf("%d\t%f\t%f\t%d\t1\t3 :\t%d\n", chunk_id++, ftimes[i], ftimes[i], fsizes[i], i);
        break;
      case 2:
        /* P Type */
        p_chunk_size += fsizes[i];
        p_chunk_frame[p_chunk_dimension++] = i;
        p_min_time = p_min_time<ftimes[i] ? p_min_time : ftimes[i];
        p_max_time = p_max_time>ftimes[i] ? p_max_time : ftimes[i];
        break;
      case 3:
        /* B Type */
        b_chunk_size += fsizes[i];
        b_chunk_frame[b_chunk_dimension++] = i;
        b_min_time = b_min_time<ftimes[i] ? b_min_time : ftimes[i];
        b_max_time = b_max_time>ftimes[i] ? b_max_time : ftimes[i];
        break;
      default:
        return;
    }
  }
}

void chunkise_gop(int n)
{
  int i, chunk_id = 0;
  int chunk_dimension = 0;
  int chunk_size = 0;
  double min_time = 0;
  double max_time = 0;

  for (i = 0; i < n; i++) {
    int j;
    int chunk_frame[150];

    if (ftypes[i] == 1) {
      /* I Type */
      if (chunk_dimension) {
        printf("%d\t%f\t%f\t%d\t1\t1 :\t", chunk_id++, min_time, max_time, chunk_size);
        for (j = 0; j < chunk_dimension; j++) {
          printf("%d ", chunk_frame[j]);
        }
        printf("\n");
      }
      chunk_size = 0;
      chunk_dimension = 0;
      min_time = DBL_MAX;
      max_time = 0;
    }
    chunk_size += fsizes[i];
    chunk_frame[chunk_dimension++] = i;
    min_time = min_time<ftimes[i] ? min_time : ftimes[i];
    max_time = max_time>ftimes[i] ? max_time : ftimes[i];
  }
}

int frames(const char* sfile){

  AVFormatContext *s;
  int done, cnt;
  int i;

  avcodec_register_all();
  av_register_all();

  s = open_input_stream(sfile);
  if (s == NULL) {
    fprintf(stderr, "Cannot open input file %s\n", sfile);

    return -1;
  }
  codec_open(s);

  done = 0; cnt = 0;
  while (!done) {
    AVPacket *pkt;

    pkt = read_input_packet(s);
    if (pkt == NULL) {
      done = 1;
    } else {
      if (s->streams[pkt->stream_index]->codec->codec_type == CODEC_TYPE_VIDEO) {
        uint8_t type;

        fsizes[cnt] = pkt->size;
        ftimes[cnt] = (double)pkt->dts/AV_TIME_BASE;
        i = pkt_decode(s, pkt, cnt, &type);
        if (i >= 0) {
          ftypes[i] = type;
        }
        cnt++;
      }
      av_free_packet(pkt);
    }
  }

  for (i = 0; i < cnt; i++) {
    fprintf(stderr,"%d\t%d %d\n", i, fsizes[i], ftypes[i]);
  }

  return cnt;
}

//usage: videofile chunksize
int main(int argc, char *argv[])
{
  int numframes = frames(argv[1]);
  
  int csize = atoi(argv[2]);
  switch (csize) {
  case -2:
    chunkise_ipb(numframes);
    break;
  case 0:
    chunkise_gop(numframes);
    break;
  default:
   //TBD support fixed sized chunks
   break;
  }

  return 0;
}
