#include "avformat.h"
#include "avcodec.h"

int codec_open(AVFormatContext *ctx)
{
  AVCodec *codec;
  int i, res;

  for (i = 0; i < ctx->nb_streams; i++) {
    codec = avcodec_find_decoder(ctx->streams[i]->codec->codec_id);
    if (!codec) {
      fprintf(stderr, "Cannot find codec %d for input stream %d\n",
                      ctx->streams[i]->codec->codec_id, i);
      return -1;
    }
    res = avcodec_open(ctx->streams[i]->codec, codec);
    if (res < 0) {
      fprintf(stderr, "Cannot open codec for input stream %d\n", i);

      return -1;
    }
  }

  return 0;
}

int pkt_decode(AVFormatContext *ctx, AVPacket *pkt, int n, uint8_t *type)
{
  AVFrame frame;
  int res, got;

  avcodec_get_frame_defaults(&frame);
  ctx->streams[pkt->stream_index]->codec->reordered_opaque= n;
  res = avcodec_decode_video(ctx->streams[pkt->stream_index]->codec,
        &frame, &got, pkt->data, pkt->size);
  if (res < 0) {
    got = res;
  }

  if (got) {
    *type = frame.pict_type;
    return frame.reordered_opaque;
  }

  return -1;
}
