/*
 *  Copyright (c) 2009 Luca Abeni
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "chunk.h"
#include "output.h"

struct out_ctx {
  int fd;
  int num_chunks;
  int num_frames;
  struct chunk *chunks;
};

static int is_full(const struct out_ctx *s, int current)
{
  int i;

  for (i = current; i < s->num_chunks; i++) {
    if (s->chunks[i].frame[0] != NULL) {
      return 1;
    }
  }

  return 0;
}

void *output_stream_open(const char *fname)
{
  struct out_ctx *s;

  s = malloc(sizeof(struct out_ctx));
  if (s == NULL) {
    return NULL;
  }

  s->chunks = NULL;
  s->num_chunks = 0;
  s->num_frames = 0;
  s->fd = open(fname, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IROTH);
  if (s->fd < 0) {
    free(s);

    return NULL;
  }

  return s;
}

int output_stream_flush(void *s_h)
{
  struct out_ctx *s = s_h;
  int i, ch, size = 0;

  for (i = 0; i < s->num_frames; i++) {
    for (ch = 0; ch < s->num_chunks; ch++) {
      if (s->chunks[ch].frame[i]) {
        size += write(s->fd, s->chunks[ch].frame[i], s->chunks[ch].frame_len[i]);
        s->chunks[ch].frame[i] = NULL;
      }
    }
  }

  return size;
}

int chunk_write(void *s_h, const struct chunk *c, int frames)
{
  struct out_ctx *s = s_h;
  int size = 0;
  int i;
  int curr_chunk = c->id % c->stride_size;
 
  if (s->num_chunks == 0) {
    s->chunks = malloc(sizeof(struct chunk) * c->stride_size);
    if (s->chunks == NULL) {
      return -1;
    }
    memset(s->chunks, 0, sizeof(struct chunk) * c->stride_size);
    s->num_chunks = c->stride_size;
  }
  if (s->num_chunks != c->stride_size) {
    return -2;
  }
  if (is_full(s, curr_chunk)) {
    size = output_stream_flush(s);
  }

  for (i = 0; i < frames; i++) {
    s->chunks[curr_chunk].frame[i] = malloc(c->frame_len[i]);
    memcpy(s->chunks[curr_chunk].frame[i], c->frame[i], c->frame_len[i]);
    s->chunks[curr_chunk].frame_len[i] = c->frame_len[i];
  }
  s->num_frames = frames;

  return size;
}

void output_stream_close(void *s_h)
{
  int *fd = s_h;

  close(*fd);
  free(fd);
}
