#!/usr/bin/perl
 ###########################################
 #  Copyright (c) 2009 Csaba Kiraly        #
 #                                         #
 #  This is free software; see gpl-3.0.txt #
 ###########################################

use strict;
use Statistics::Descriptive;
use POSIX qw(ceil floor);
use Switch;

my $gl_inputfolder;

($ENV{'PSNR_TOOLS'}) or (print "\$PSNR_TOOLS not defined\n" and die);
my $PSNR_TOOLS=$ENV{'PSNR_TOOLS'};
($ENV{'PSNR_TMPPREFIX'}) or (print "\$PSNR_TMPPREFIX not defined\n" and die);
my $PSNR_TMPPREFIX=$ENV{'PSNR_TMPPREFIX'};

#a value of 'src' for the peer creates an empty lost file
sub peer_chunks_lost ($) {
  my ($peer)=@_;

  my @losts;
  if ($peer=='src') {
    @losts=`echo -n > lost.txt`;
  } else {
    @losts=`$PSNR_TOOLS/chunks_lost.sh $peer $gl_inputfolder| tee lost.txt`;
  }
  chomp @losts;
  return @losts;
}

sub peer_qoe_calc ($$) {
  my ($peer,$chunksize)=@_;

  my @losts = peer_chunks_lost($peer);


  return `$PSNR_TOOLS/psnr.sh calc $chunksize lost.txt`;
}

sub peer_qoe_process ($) {
  my ($calc_result)=@_;

  my %qoe;

  $calc_result =~ m/psnr: ([\d\.]+)/;
  $qoe{ 'psnr' } = $1;

  $calc_result =~ m/ssim: ([\d\.]+)/;
  $qoe{ 'ssim' } = $1;

  $calc_result =~ m/bytes: ([\d\.]+)/;
  $qoe{ 'bytes' } = $1;

  $calc_result =~ m/chunks: ([\d\.]+)/;
  $qoe{ 'chunks' } = $1;

  $calc_result =~ m/chunkslost: ([\d\.]+)/;
  $qoe{ 'chunkslost' } = $1;

  $qoe{ 'chunkloss' } = $qoe{ 'chunkslost' } / $qoe{ 'chunks' } ;

  return \%qoe;
}

sub peer_qoe($$) {
  my ($peer,$chunksize)=@_;
  my $q = peer_qoe_calc ($peer,$chunksize);
  return peer_qoe_process($q);
}

#sub num_chunks() {
#  return 1000;
#}
#
#sub peer_chunkloss ($@) {
#  my ($peer,@losts)=@_;
#  return (scalar @losts) / num_chunks();
#}

sub peers_of_class ($) {
  my ($peerclass)=@_;

  my @peers=`$PSNR_TOOLS/peers_of_class.sh $peerclass $gl_inputfolder`;
  chomp @peers;
  return @peers;
}

sub print_stat($\%$) {
  my ($name,$qoes,$metric)=@_;

  my $stat = Statistics::Descriptive::Full->new();
  foreach my $peer (keys(%$qoes)) {
    $stat->add_data($qoes->{$peer}{$metric});
  }

  print "$name: ";
  print " mean=".$stat->mean(); 
  print " var=".$stat->variance();
  print " count=".$stat->count();
  print "\n";

  return $stat->mean();
}

sub print_stats($\%){
  my ($peerclass,$qoes)=@_;
  my %avgqoe;
  $avgqoe{'psnr'} = print_stat("  psnr(uncompressed->received)",%$qoes ,'psnr');
  $avgqoe{'ssim'} = print_stat("  ssim",%$qoes,'ssim');
  $avgqoe{'chunkloss'} = print_stat("  chunkloss",%$qoes,'chunkloss');

  print "PSNRresults $peerclass $avgqoe{'psnr'} $avgqoe{'ssim'}\n";
  print "\n";

  return \%avgqoe;
}

sub qoe($$){
  my ($chunksize,$every)=@_;

  my %allqoes;
  my @peerclasses=(1,2,3,4);

  foreach my $peerclass (@peerclasses) {

    print " Peer class = $peerclass\n";

    my @peers=peers_of_class($peerclass);

    my %qoes;
    my $everycounter=0;
    foreach my $peer (@peers) {
      if ($everycounter % $every == 0) {
        $qoes{$peer} = peer_qoe($peer,$chunksize);
      }
      $everycounter++;
    }
    print_stats($peerclass,%qoes);
    @allqoes{keys %qoes} = values %qoes;
  }

  print " All peers\n";
  my $avgqoe = print_stats("all",%allqoes);

  return $avgqoe;
}

sub init ($$) {
  my ($video,$coder)=@_;
  my $tmp=`$PSNR_TOOLS/psnr.sh init $video "$coder"`;

  my $srcqoe = peer_qoe('src',1024); # it doesen't matter what chunksize we use, since it calculates without losses
  print "psnr(uncompressed->compressed): ".$srcqoe->{'psnr'}."\n";
  print "ssim(uncompressed->compressed): ".$srcqoe->{'ssim'}."\n";
  print "PSNRencoded $srcqoe->{'psnr'} $srcqoe->{'ssim'}\n";
  print "\n";
  #my $chunksize = ceil($srcqoe->{'bytes'}/$srcqoe->{'chunks'}); 
}

sub chunk ($) {
  my ($chunksize)=@_;
  my $tmp=`$PSNR_TOOLS/psnr.sh chunk $chunksize`;
}

sub calc ($$$) {
  my ($inputfolder,$chunksize,$every)=@_;
  $gl_inputfolder=$inputfolder;

#  print "\nQuality with GOP size chunks (avg chunk size = $srcqoe->{'bytes'}/$srcqoe->{'chunks'} bytes )\n";
#  my $avgqoe_gop = qoe(0,$every);
#  print "\nQuality with $chunksize bytes chunks\n";
#  my $avgqoe_nogop = qoe($chunksize,$every);
#  print "\nQuality with 2-3 chunks grouping I,P,B frmaes per GOP\n";
#  my $avgqoe_ipb = qoe(-2,$every);
  qoe($chunksize,$every);
}

#--------------------------------------------------

if ($ARGV[0] eq 'init') {
  ($#ARGV + 1 == 3) or die;
  my $video=$ARGV[1];
  my $coder=$ARGV[2];
  init($video,$coder);
} elsif ($ARGV[0] eq 'chunk') {
  ($#ARGV + 1 == 2) or die;
  my $chunksize=$ARGV[1];
  chunk($chunksize);
} elsif ($ARGV[0] eq 'calc') {
  ($#ARGV + 1 == 4) or die;
  my $inputfolder=$ARGV[1];
  my $chunksize=$ARGV[2];
  my $every=$ARGV[3]; #calculate values only for every "everty"th of peers
  calc($inputfolder,$chunksize,$every);
} else {
  print "Usage: ... see source\n";
}

