/*
 *  Copyright (c) 2009 Luca Abeni
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <stdlib.h>

//#include <libavdevice/avdevice.h>
#include <libavformat/avformat.h>

#include "codec.h"

void codec_init(void)
{
    avcodec_register_all();
//    avdevice_register_all();
    av_register_all();

}

int codec_is_audio(void *h, int i)
{
    AVFormatContext *s = h;

    return (s->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO); 
}

int stream_copy(void *h, int i)
{
    AVFormatContext *s = h;

    return (s->streams[i]->stream_copy == 1); 
}

int codec_is_video(void *h, int i)
{
    AVFormatContext *s = h;

    return (s->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO); 
}

int num_streams(void *h)
{
    AVFormatContext *s = h;

    return s->nb_streams;
}

void index_set(void *h, int i)
{
  AVPacket *pkt = h;

  pkt->stream_index = i;
}

void ts_copy(void *dst_h, void *out_h, void *src_h, void *in_h, int id)
{
  AVFormatContext *out = out_h;
  AVFormatContext *in = in_h;
  AVPacket *dst = dst_h;
  AVPacket *src = src_h;

  if (src->pts != AV_NOPTS_VALUE) {
    dst->pts = av_rescale_q(src->pts, in->streams[id]->time_base, out->streams[dst->stream_index]->time_base);
  } else {
    dst->pts = AV_NOPTS_VALUE;
  }
  if (src->dts != AV_NOPTS_VALUE) {
    dst->dts = av_rescale_q(src->dts, in->streams[id]->time_base, out->streams[dst->stream_index]->time_base);
  } else {
    dst->dts = AV_NOPTS_VALUE;
  }
  dst->duration = av_rescale_q(src->duration, in->streams[id]->time_base, out->streams[dst->stream_index]->time_base);
}

void pkt_stream_assign(void *h, int i)
{
  AVPacket *pkt = h;

  pkt->stream_index = i;
}

int pkt_stream(void *h)
{
  AVPacket *pkt = h;

  return pkt->stream_index;
}

void *encode(void *s_h, void *f_h, int i)
{
  static AVPacket p;
  static void *coded_buffer;
  static int buffer_size;
  AVFrame *f = f_h;
  AVFormatContext *s = s_h;
  AVCodecContext *c = s->streams[i]->codec;
  int res;

  if (coded_buffer == NULL) {
    buffer_size = 4 * c->width * c->height;
    if (buffer_size < 1024 * 1024) buffer_size = 1024 * 1024;
    coded_buffer = malloc(buffer_size);
  }
  res = avcodec_encode_video(c, coded_buffer, buffer_size, f_h);
  av_init_packet(&p);
  p.stream_index = i;
  p.data= coded_buffer;
  p.size= res;
  p.pts = f->reordered_opaque;
  if(c->coded_frame && c->coded_frame->key_frame) {
    p.flags |= PKT_FLAG_KEY;
  }

  return &p;
}

void *frame_copy(void *h, int id, void *f_h)
{
  int i;
  static AVFrame frame, frame_template;
  AVFrame *f = f_h;
  AVFormatContext *s = h;
  AVCodecContext *c = s->streams[id]->codec;

  if (frame_template.data[0] == NULL) {
    avcodec_get_frame_defaults(&frame_template);
    avpicture_alloc((AVPicture *)&frame_template, c->pix_fmt, c->width, c->height);
  }

  frame = *f;
  for (i = 0; i < 4; i++) {
    frame.data[i] = frame_template.data[i];
    frame.linesize[i] = frame_template.linesize[i];
  }
  av_picture_copy(&frame, f, c->pix_fmt, c->width, c->height);

  return &frame;
}

void *decode(void *s_h, void *p_h)
{
  int res, len, got;
  void *ptr;
  static AVFrame pict;
  AVPacket *pkt = p_h;
  AVFormatContext *s = s_h;
  AVCodecContext *c = s->streams[pkt->stream_index]->codec;
  
  len = pkt->size;
  ptr = pkt->data;

  avcodec_get_frame_defaults(&pict);
  c->reordered_opaque = pkt->pts;
  res = avcodec_decode_video(c, &pict, &got, ptr, len);
  if ((res < 0) || (got == 0)) {
    return NULL;
  }
  //pict.pict_type = 0;
  //pict.pts = AV_NOPTS_VALUE;
 
  return &pict;
}

void frame_pts_increase(void *f_h, int delta)
{
  AVFrame *f = f_h;

  f->reordered_opaque += delta;
}

uint64_t frame_pts(void *f_h)
{
  AVFrame *f = f_h;

  return f->reordered_opaque;
}

void frame_pts_null(void *f_h)
{
  AVFrame *f = f_h;

  f->pts = AV_NOPTS_VALUE;
  f->reordered_opaque = AV_NOPTS_VALUE;
}

#if 0

static AVFrame accum;

/* FAKE! */
void *accumulate(void *s_h, void *s_f, int i)
{
  AVFormatContext *s = s_h;
  AVCodecContext *c = s->streams[1]->codec;
  AVFrame *f = s_f;
  static int inited;
  int j, x, y;
  
  if (!inited) {
    inited = 1;

    avpicture_alloc(&accum, c->pix_fmt, c->width, c->height);
  }
 
  if (i == 1) {
    av_picture_copy(&accum, f, c->pix_fmt, c->width, c->height);

    accum.key_frame = f->key_frame;
    accum.pict_type = f->pict_type;
    accum.pts = f->pts;
    accum.coded_picture_number = f->coded_picture_number;
    accum.display_picture_number = f->display_picture_number;
    accum.quality = f->quality;
    return &accum;
  }
  
  for (j = 0; j < 3; j++) {
    uint8_t *dst = accum.data[j];
    uint8_t *src = f->data[j];

    for (y = 0; y < s->streams[1]->codec->height >> (j != 0); y++) {
      for (x = 0; x < s->streams[1]->codec->width >> (j != 0); x++) {
        dst[x + y * accum.linesize[j]] = ((int)dst[x + y * accum.linesize[j]] + (int)src[x + y * c->coded_frame->linesize[j]] - 127);
      }
    }
  }

  return &accum;
}
#endif
