#!/bin/bash
 ###########################################
 #  Copyright (c) 2009 Csaba Kiraly        #
 #                                         #
 #  This is free software; see gpl-3.0.txt #
 ###########################################

[[ $PSNR_TOOLS ]] || { echo "\$PSNR_TOOLS not defined"; exit 1; }
[[ $PSNR_TMPPREFIX ]] || { echo "\$PSNR_TMPPREFIX not defined"; exit 1; }
[[ $FFMPEG ]] || { echo "\$FFMPEG not defined"; exit 1; }

mkdir -p `dirname ${PSNR_TMPPREFIX}_`

ORIG_YUV=${PSNR_TMPPREFIX}orig.yuv
ORIG_YUV_DESC=${ORIG_YUV}.desc
ORIG_MP4=${PSNR_TMPPREFIX}orig_compressed.m4v
RECEIVED_MP4=${PSNR_TMPPREFIX}received.m4v
REFILLED_YUV=${PSNR_TMPPREFIX}refilled.yuv

ANALYSER=$PSNR_TOOLS/Analyser/analyser
CHUNKER=$PSNR_TOOLS/Chunker/chunker
CUTTER=$PSNR_TOOLS/Cutter/cutter
LOSTFRAMES=$PSNR_TOOLS/LostFrames/lf
LOGPARSER=$PSNR_TOOLS/LogParser/logparser
REFILLER=$PSNR_TOOLS/ReFiller/refiller
EVALPSNR=$PSNR_TOOLS/EvalVid/psnr


function usage {
   echo "Usage: $0 init <orig video> <codec settings>";
   echo "       $0 chunk <chunksize>";
   echo "       $0 calc <chunksize> <file with losses>";
   echo "       $0 play <chunksize> <file with losses>";
   echo "examples:";
   echo " $0 init ~/video/foreman_cif_100GOP.yuv '-f h264 -vcodec libx264 -vpre libx264-default -keyint_min 12  -g 12 -b 0.4M '";
   echo " $0 calc 0 lost.txt";
   echo " $0 play 1500 lost.txt";
   echo "";
   exit $1 
}

function init {
  [[ $# -eq 2 ]] || usage 1

  ORIG=$1 # the original sequence. Can be a YUV but also a something encoded. If it is a YUV, you need a .desc file as well!!!
  ORIG_DESC=${ORIG}.desc
  CODER=$2 # conversion parameters for streaming

  #initialize by creating YUV, if needed
  rm $ORIG_YUV $ORIG_YUV_DESC
  if [[ $ORIG =~ \.[yY][uU][vV]$ ]]; then #if it is aready a YUV, just link or copy
    ln -s  $ORIG $ORIG_YUV || cp $ORIG $ORIG_YUV
    ln -s  $ORIG_DESC $ORIG_YUV_DESC || cp $ORIG_DESC $ORIG_YUV_DESC || { echo "$ORIG_DESC missing. You need a description file for a YUV"; exit 1; }
    . $ORIG_YUV_DESC
    [[ $VIDEO_XSIZE != "" && $VIDEO_YSIZE != "" && $VIDEO_FPS != "" ]] || { echo "$ORIG_YUV_DESC not valid. You need a description file for a YUV"; exit 1; }

  else
    #$FFMPEG -i $ORIG -s ${VIDEO_XSIZE}x${VIDEO_YSIZE} -r $VIDEO_FPS $ORIG_YUV >${PSNR_TMPPREFIX}ffmpeg.log
    $FFMPEG -y -i $ORIG $ORIG_YUV >${PSNR_TMPPREFIX}ffmpeg.log 2>&1

    # Stream #0.0: Video: h264, yuv420p, 352x288, 25 tbr, 1200k tbn, 50 tbc
    TMP=`grep "Stream.*Video:" ${PSNR_TMPPREFIX}ffmpeg.log`
#    if [[ $TMP =~ Stream.*Video:.*\ ([0-9]+)x([0-9]+),\ ([0-9]+)\ tbr ]]; then #working on GNU bash, version 3.2.48(1)-release (x86_64-pc-linux-gnu)
    if [[ $TMP =~ Stream.*Video:.*\ \([0-9]+\)x\([0-9]+\),\ \([0-9]+\)\ tbr ]]; then #working on GNU bash, version 3.1.17(1)-release (x86_64-pc-linux-gnu), also on version 3.2.48(1)-release (x86_64-pc-linux-gnu)
      VIDEO_XSIZE=${BASH_REMATCH[1]}
      VIDEO_YSIZE=${BASH_REMATCH[2]}
      VIDEO_FPS=${BASH_REMATCH[3]}
      echo "VIDEO_XSIZE=$VIDEO_XSIZE"  >$ORIG_YUV_DESC
      echo "VIDEO_YSIZE=$VIDEO_YSIZE" >>$ORIG_YUV_DESC
      echo "VIDEO_FPS=$VIDEO_FPS" >>$ORIG_YUV_DESC
    else
      echo "Can't get video description (XSIZE, YSIZE or FPS). You need to create a description file by hand"
      exit 1
    fi
  fi

  #$FFMPEG -s ${VIDEO_XSIZE}x${VIDEO_YSIZE} -r 25 -i $ORIG_YUV -f m4v -mbd 2 -b 400k $ORIG_MP4 >>${PSNR_TMPPREFIX}ffmpeg.log
  $FFMPEG -y -s ${VIDEO_XSIZE}x${VIDEO_YSIZE} -r $VIDEO_FPS -i $ORIG_YUV $CODER $ORIG_MP4 >>${PSNR_TMPPREFIX}ffmpeg.log
  $ANALYSER $ORIG_MP4 > ${PSNR_TMPPREFIX}orig_log.txt
}

function chunk {
  [[ $# -eq 1 ]] || usage 1
  CHUNK_SIZE=$1
  $CHUNKER $ORIG_MP4 $CHUNK_SIZE  > ${PSNR_TMPPREFIX}chunks.txt
}

function calc {
  [[ $# -eq 2 ]] || usage 1

  CHUNK_SIZE=$1
  LOST_FILE=$2

  . $ORIG_YUV_DESC

  #cut lost chunks
  rm $RECEIVED_MP4 # FIXME: I have no idea why this was needed, but sometimes cutter wasn't overwriting
  if [[ $CHUNK_SIZE -eq -2 || $CHUNK_SIZE -eq 0  ]]; then
    $LOSTFRAMES ${PSNR_TMPPREFIX}chunks.txt $LOST_FILE > ${PSNR_TMPPREFIX}lostframes.txt
    $CUTTER $ORIG_MP4 $RECEIVED_MP4 ${PSNR_TMPPREFIX}lostframes.txt $CHUNK_SIZE 2> ${PSNR_TMPPREFIX}cutter_log.txt
  else
    $CUTTER $ORIG_MP4 $RECEIVED_MP4 $LOST_FILE $CHUNK_SIZE 2> ${PSNR_TMPPREFIX}cutter_log.txt
  fi

  #get chunk count and avgsize
  BYTES=`awk '/^bytes:/ {print $2}' ${PSNR_TMPPREFIX}cutter_log.txt`
  CHUNKS=`awk '/^chunks:/ {print $2}' ${PSNR_TMPPREFIX}cutter_log.txt`
  CHUNKSLOST=`awk '/^chunkslost:/ {print $2}' ${PSNR_TMPPREFIX}cutter_log.txt`
  #CHUNK_SIZE_AVG=`(echo - | awk '{ print $BYTES/$CHUNKS}')`

  #analyse losses
  $ANALYSER $RECEIVED_MP4 > ${PSNR_TMPPREFIX}received_log.txt

  #reconstruct sequence by duplicating 
  $LOGPARSER ${PSNR_TMPPREFIX}orig_log.txt ${PSNR_TMPPREFIX}received_log.txt > ${PSNR_TMPPREFIX}received_dup.txt
  $REFILLER $RECEIVED_MP4 $REFILLED_YUV ${PSNR_TMPPREFIX}received_dup.txt > ${PSNR_TMPPREFIX}refiller_log.txt
  $EVALPSNR $VIDEO_XSIZE $VIDEO_YSIZE 420 $ORIG_YUV $REFILLED_YUV >${PSNR_TMPPREFIX}psnr_log.txt 2>${PSNR_TMPPREFIX}psnravg_log.txt

  PSNR=`awk '/^psnr:/ {print $8}' ${PSNR_TMPPREFIX}psnravg_log.txt`

  $EVALPSNR $VIDEO_XSIZE $VIDEO_YSIZE 420 $ORIG_YUV $REFILLED_YUV ssim >${PSNR_TMPPREFIX}ssim_log.txt 2>${PSNR_TMPPREFIX}ssimavg_log.txt

  SSIM=`awk '/^ssim:/ {print $8}' ${PSNR_TMPPREFIX}ssimavg_log.txt`

  echo "bytes: $BYTES"
  echo "chunks: $CHUNKS"
  echo "chunkslost: $CHUNKSLOST"
  echo "frames: $FRAMES"
  echo "frameslost: $FRAMESLOST"
  echo "psnr: $PSNR"
  echo "ssim: $SSIM"
}

function play {
  [[ $# -eq 2 ]] || usage 1
  [[ $FFPLAY ]] || { echo "\$FFPLAY not defined"; exit 1; }

  #run all the calculations
  calc "$@"

  #play the resulting video
  #$FFPLAY -s ${VIDEO_XSIZE}x${VIDEO_YSIZE} -r $VIDEO_FPS -i $ORIG_YUV
  $FFPLAY -s ${VIDEO_XSIZE}x${VIDEO_YSIZE} $REFILLED_YUV
}

[[ $# -ge 1 ]] || usage 1
CMD=$1;
shift

case $CMD in
  init)
    init "$@"
    ;;
  chunk)
    chunk "$@"
    ;;
  calc)
    calc "$@"
    ;;
  play)
    play "$@"
    ;;
  *)
    usage 1
    ;;
esac
